(declaim (optimize (speed 3)
		   (compilation-speed 0)
		   (safety 1)
		   (debug 0)))

#||
(defmacro debout (&rest args)
  `(progn
     (format *error-output* ,@args)
     (force-output *error-output*)))
||#

(defmacro debout (&rest args)
  (declare (ignore args))
  nil)


(defstruct dll
  (prv nil :type (or null dll))
  (nxt nil :type (or null dll))
  (datum nil))

(defun make-dll-head ()
  (let ((ret (make-dll)))
    (setf (dll-prv ret) ret)
    (setf (dll-nxt ret) ret)
    ret))

(defmacro dll-empty (head)
  (let ((hd (gensym)))
    `(let ((,hd ,head))
       (eq (dll-nxt ,hd) ,hd))))

(defun dll-next (hd el)
  (let ((ret (dll-nxt el)))
    (if (eq ret hd)
	nil
      ret)))

(defun dll-prev (hd el)
  (let ((ret (dll-prv el)))
    (if (eq ret hd)
	nil
      ret)))

(defun dll-push (hd el)
  (let* ((nxt (dll-nxt hd))
	 (new (make-dll :datum el :prv hd :nxt nxt)))
    (system:without-interrupts
     (setf (dll-nxt hd) new)
     (setf (dll-prv nxt) new)))
  hd)

(defun dll-append (hd el)
  (let* ((prv (dll-prv hd))
	 (new (make-dll :datum el :prv prv :nxt hd)))
    (system:without-interrupts
     (setf (dll-prv hd) new)
     (setf (dll-nxt prv) new))
    hd))

(defun dll-remove (el)
  (let ((nxt (dll-nxt el))
	(prv (dll-prv el)))
    (system:without-interrupts
     (setf (dll-prv nxt) prv)
     (setf (dll-nxt prv) nxt))
    nil))

(defmacro dll-do ((el dll &optional ret) &rest body)
  (let ((lst (gensym))
	(nxt (gensym)))
    `(do* ((,lst ,dll)
	   (,el (dll-nxt ,lst) (dll-nxt ,nxt))
	   (,nxt ,el ,el))
	 ((eq ,el ,lst) ,ret)
       ,@body)))

(defmacro dll-first (head)
  (let ((hd (gensym)))
    `(let ((,hd ,head))
       (dll-next ,hd ,hd))))

(defmacro dll-last (head)
  (let ((hd (gensym)))
    `(let ((,hd ,head))
       (dll-prev ,hd ,hd))))

(defun dll-repl-dll (node dll)
  (let ((prv (dll-prv node))
	(nxt (dll-nxt node))
	(fst (dll-first dll))
	(lst (dll-last dll)))
    (if fst
	(system:without-interrupts
	 (setf (dll-nxt prv) fst)
	 (setf (dll-prv fst) prv)
	 (setf (dll-prv nxt) lst)
	 (setf (dll-nxt lst) nxt)
	 nil)
      (dll-remove node))))

(defun dll-clear (hd)
  (system:without-interrupts
   (setf (dll-nxt hd) hd)
   (setf (dll-prv hd) hd)))

(defun dll-repl-all (hd el)
  (let ((node (make-dll :prv hd :nxt hd :datum el)))
    (system:without-interrupts
     (setf (dll-nxt hd) node)
     (setf (dll-prv hd) node))))

(defun dll-concat-strings (dll)
  (let ((total-len 0))
    (dll-do (node dll)
      (let ((el (dll-datum node)))
	(declare (type simple-base-string el))
	(incf total-len (length el))))
    (if (zerop total-len)
	(dll-clear dll)
      (let ((all (make-string total-len))
	    (start 0))
	(dll-do (node dll)
	  (let ((el (dll-datum node)))
	    (declare (type simple-base-string el))
	    (let ((len (length el)))
	      (setf (subseq all start (+ start len)) el)
	      (incf start len))))
	(dll-repl-all dll all)))))

(defparameter *best-so-far* nil)
(defparameter *orig-input* nil)

(defun load-timer (arg)
  (multiple-value-bind (fine errno)
      (unix:unix-setitimer :real 0 0 arg 0)
    (unless fine
      (warn "Can't set timer: ~A" (unix:get-unix-error-msg errno)))))

(defun get-input (str)
  (let* ((size #x100000)
	 (ret (make-string size))
	 (write-pos (read-sequence ret str)))
    (if (< write-pos size)
	(subseq ret 0 write-pos)
      (concatenate 'string ret (get-input str)))))

(defparameter *token-table* (make-token-table))
(declaim (type (simple-array (or null token) (256 256)) *token-table*))

(defparameter *space-table* (make-space-tab))
(declaim (type (simple-base-string 256) *space-table*))

(defmacro isspace (c)
  `(eq (schar *space-table* (char-code ,c)) #\Space))

(defun get-token (file flen start)
  (declare (fixnum flen start)
	   (type simple-base-string file))
  (flet ((read-spec (sstart)
	   (if (< sstart (1- flen))
	       (let ((token (aref *token-table*
				  (char-code (schar file sstart))
				  (char-code (schar file (1+ sstart))))))
		 (if (and token
			  (or (= 1 (token-len token))
			      (and (< (+ sstart 2) flen)
				   (eq (schar file (+ sstart 2)) #\>))))
		     token
		   (error "Invalid file")))
	     (error "Invalid file"))))
    (if (>= start flen)
	'eof
      (case (schar file start)
	(#\< (if (eq (schar file (1+ start)) #\/)
		 (let ((token (read-spec (+ start 2))))
		   (values 'end (token-type token) (token-val token)
			   (the fixnum (+ start (token-len token) 3))))
	       (let ((token (read-spec (1+ start))))
		 (values 'start (token-type token) (token-val token)
			 (the fixnum (+ start (token-len token) 2))))))
	(#\> (error "Invalid file"))
	(otherwise (do ((i (1+ start) (1+ i)))
		       ((or (>= i flen) (eq (schar file i) #\<))
			(values 'text 'string
				(subseq file start i) i))
		     (declare (fixnum i))))))))

(defstruct region
  (type 'size :type (member attr size color))
  (val 0 :type (or fixnum symbol character))
  (region nil :type (or null dll)))

(defun do-parse (file flen)
  (declare (fixnum flen)
	   (type simple-base-string file))
  (labels
      ((parse-until (sstart fin-tok fin-type fin-val acc)
	 (multiple-value-bind (token type val next)
	     (get-token file flen sstart)
	   (if (and (eq token fin-tok) (eq type fin-type)
		    (eq val fin-val))
	       (values acc next)
	     (ecase token
	       (start (multiple-value-bind (new-acc new-next)
			  (parse-until next 'end type val (make-dll-head))
			(parse-until new-next fin-tok fin-type fin-val
				     (if (not (dll-empty new-acc))
					 (dll-append acc
						     (make-region :type type
								  :val val
								  :region
								  new-acc))
				       acc))))
	       (text (parse-until next fin-tok fin-type fin-val
				  (dll-append acc val)))
	       ((end eof) (error "Invalid file")))))))
    (parse-until 0 'eof nil nil (make-dll-head))))

(defun token-chars (type val)
  (ecase type
    (color val)
    (size (code-char (+ val (char-code #\0))))
    (attr (ecase val
	    (bold #\B)
	    (emph (values #\E #\M))
	    (italic #\I)
	    (plain (values #\P #\L))
	    (strong #\S)
	    (tt (values #\T #\T))
	    (uline #\U)))))

(defun dump (tree)
  (dolist (hd tree)
    (etypecase hd
      (simple-base-string (write-string hd))
      (region (multiple-value-bind (fst snd)
		  (token-chars (region-type hd) (region-val hd))
		(write-char #\<)
		(write-char fst)
		(when snd (write-char snd))
		(write-char #\>)
		(dump (region-region hd))
		(write-string "</")
		(write-char fst)
		(when snd (write-char snd))
		(write-char #\>)))))
  (finish-output))

(defun dump (tree)
  (dll-do (el tree)
    (let ((hd (dll-datum el)))
      (etypecase hd
	(simple-base-string (write-string hd))
	(region (multiple-value-bind (fst snd)
		    (token-chars (region-type hd) (region-val hd))
		  (write-char #\<)
		  (write-char fst)
		  (when snd (write-char snd))
		  (write-char #\>)
		  (dump (region-region hd))
		  (write-string "</")
		  (write-char fst)
		  (when snd (write-char snd))
		  (write-char #\>)))))
    (finish-output)))

(defun save-day ()
  (cond (*best-so-far* (dump *best-so-far*))
	(*orig-input* (write-string *orig-input*))
	(t (error "Couldn't even read input!")))
  (finish-output)
  (unix:unix-exit 0))

(defstruct props
  (b nil :type (member t nil))
  (em nil :type (member t nil))
  (i nil :type (member t nil))
  (s nil :type (member t nil))
  (tt nil :type (member t nil))
  (u 0 :type (integer 0 3))
  (size -1 :type (integer -1 9)) ; -1 means unknown
  (color #\x :type (member #\r #\g #\b #\c #\m #\y #\k #\w #\x))) ; x unknown

(defun update-props (old type val)
  (macrolet
    ((update-if-new (el)
       (let ((x (gensym)) (oldv (gensym)))
	 `(let ((,oldv (,el old)))
	    (if (eq ,oldv val)
		nil
	      (progn
		(setf (,el old) val)
		(lambda (,x)
		  (declare (type props ,x))
		  (setf (,el ,x) ,oldv)
		  ,x))))))
     (update-if-not (el)
       (let ((x (gensym)))
	 `(if (,el old)
	      nil
	    (progn
	      (setf (,el old) t)
	      (lambda (,x)
		(declare (type props ,x))
		(setf (,el ,x) nil)
		,x)))))
     (update-if-some (uform &rest bforms)
       (let ((oldu (gensym))
	     (x (gensym))
	     (vars (mapcar (lambda (x)
			     (declare (ignore x))
			     (gensym))
			   bforms)))
	 (labels ((compute-ifs (vals bools)
		    (if vals
			`(if ,(car vals)
			     ,(compute-ifs (cdr vals) (cons t bools))
			   ,(compute-ifs (cdr vals) (cons nil bools)))
		      (let ((pforms (remove nil
					    (mapcar (lambda (bo pr)
						      (when bo
							`(setf (,pr ,x) t)))
						    bools
						    (reverse bforms))))
			    (sforms (remove nil
					    (mapcar (lambda (bo pr)
						      (when bo
							`(setf (,pr old) nil)))
						    bools
						    (reverse bforms)))))
			(if pforms
			    `(if (zerop ,oldu)
				 (progn
				   ,@sforms
				   (lambda (,x)
				     (declare (type props ,x))
				     ,@pforms
				     ,x))
			       (progn
				 (setf (,uform old) 0)
				 ,@sforms
				 (lambda (,x)
				   (declare (type props ,x))
				   ,@pforms
				   (setf (,uform ,x) ,oldu)
				   ,x)))
			  `(if (zerop ,oldu)
			       nil
			     (progn
			       (setf (,uform old) 0)
			       (lambda (,x)
				 (declare (type props ,x))
				 (setf (,uform ,x) ,oldu)
				 ,x))))))))
	   `(let ((,oldu (,uform old))
		  ,@(mapcar (lambda (x y)
			      `(,x (,y old)))
			    vars bforms))
	      ,(compute-ifs vars nil))))))

    (ecase type
      (size (update-if-new props-size))
      (color (update-if-new props-color))
      (attr (ecase val
	      (bold (update-if-not props-b))
	      (italic (update-if-not props-i))
	      (tt (update-if-not props-tt))
	      (plain (update-if-some props-u
				     props-s props-em props-i
				     props-b props-tt))
	      (strong (if (props-s old)
			  nil
			(let ((oldem (props-em old)))
			  (setf (props-s old) t)
			  (if oldem
			      (progn
				(setf (props-em old) nil)
				(lambda (x)
				  (declare (type props x))
				  (setf (props-em x) t)
				  (setf (props-s x) nil)
				  x))
			    (lambda (x)
			      (declare (type props x))
			      (setf (props-s x) nil)
			      x)))))
	      (emph (if (props-s old)
			nil
		      (progn
			(setf (props-em old) (not (props-em old)))
			(lambda (x)
			  (declare (type props x))
			  (setf (props-em x) (not (props-em x)))
			  x))))
	      (uline (let ((oldv (props-u old)))
		       (if (>= oldv 3)
			   nil
			 (progn
			   (incf (props-u old))
			   (lambda (x)
			     (declare (type props x))
			     (decf (props-u x))
			     x))))))))))

(defun remove-redundancy (tree props)
  (dll-do (node tree)
    (let ((el (dll-datum node)))
      (etypecase el
	(string nil)
	(region (let ((f (update-props props
				       (region-type el) (region-val el))))
		  (declare (type (or null function) f))
		  (remove-redundancy (region-region el) props)
		  (if f
		      (funcall f props)
		    (dll-repl-dll node (region-region el)))))))))

(defun save-space-props (oldprops newprops)
  (declare (type props oldprops newprops))
  (setf (props-size oldprops) (props-size newprops))
  (let ((newu (props-u newprops)))
    (setf (props-color oldprops) (if (zerop newu)
				     #\w
				   (props-color newprops)))
    (setf (props-u oldprops) newu))
  (setf (props-tt oldprops) (props-tt newprops)))

(defun same-space-props (oldprops newprops)
  (and (eq (props-size oldprops) (props-size newprops))
       (let ((newu (props-u newprops)))
	 (and (eq (props-u oldprops) newu)
	      (eq (props-color oldprops)
		  (if (zerop newu)
		      #\w
		    newu))))
       (eq (props-tt oldprops) (props-tt newprops))))

(defun reduce-whitespace (tree)
  (let ((oldprops (make-props))
	(newprops (make-props))
	(last-was-space nil)
	(buf (make-array 10000 :element-type 'base-char
			 :initial-element #\Space
			 :adjustable t
			 :fill-pointer 0)))
    (labels ((fix-string (s)
	       (declare (type simple-base-string s))
	       (let ((len (length s))
		     (is-different nil)
		     (only-space t))
		 (declare (fixnum len))
		 (if (zerop len)
		     (values s only-space)
		   (let ((need-copy t)
			 (fst (schar s 0)))
		     (setf (fill-pointer buf) 0)
		     (if (isspace fst)
			 (if (and last-was-space (same-space-props oldprops
								   newprops))
			     (setq need-copy nil
				   is-different t)
			   (progn
			     (vector-push-extend #\Space buf)
			     (setf last-was-space t)))
		       (progn
			 (setf only-space nil)
			 (vector-push-extend fst buf)
			 (setf last-was-space nil)))
		     (loop for i fixnum from 1 below len do
			   (let ((chr (schar s i)))
			     (if (isspace chr)
				 (if last-was-space
				     (setf is-different t)
				   (progn
				     (vector-push-extend #\Space buf)
				     (setf last-was-space t)))
			       (progn
				 (setf only-space nil)
				 (vector-push-extend chr buf)
				 (setf last-was-space nil)))))
		     (when (and last-was-space need-copy)
		       (save-space-props oldprops newprops))
		     (if is-different
			 (values (copy-seq buf) only-space)
		       (values s only-space))))))
	     (do-reduce (dll)
	       (let ((only-space t))
		 (dll-do (node dll only-space)
		   (let ((el (dll-datum node)))
		     (etypecase el
		       (simple-base-string
			       (if (props-tt newprops)
				   (progn
				     (dotimes (i (length el))
				       (when (not (isspace (schar el i)))
					 (setf only-space nil)
					 (return)))
				     (if (eq (schar el (1- (length el)))
					     #\Space)
					 (progn
					   (save-space-props oldprops newprops)
					   (setf last-was-space t))
				       (setf last-was-space nil)))
				 (multiple-value-bind (new-string
						       new-only-space)
				     (fix-string el)
				   (unless new-only-space
				     (setf only-space nil))
				   (if (zerop (length new-string))
				       (dll-remove node)
				     (setf (dll-datum node) new-string)))))
		       (region (let ((f (update-props newprops
						      (region-type el)
						      (region-val el))))
				 (declare (type (or null function) f))
				 (if (do-reduce (region-region el))
				     (if (and (eq (region-type el) 'attr)
					      (find (region-val el)
						    '(strong emph
						      italic bold)
						    :test #'eq))
					 (progn
					   (if (dll-empty (region-region el))
					       (dll-remove node)
					     (dll-repl-dll node
						    (region-region el)))
					   (when f (funcall f newprops)))
				       (progn
					 (when f (funcall f newprops))
					 (if (dll-empty (region-region el))
					     (dll-remove node)
					   (unless f
					     (dll-repl-dll node
						  (region-region el))))))
				   (progn
				     (setf only-space nil)
				     (when f (funcall f newprops))
				     (if (dll-empty (region-region el))
					 (dll-remove node)
				       (unless f
					 (dll-repl-dll node
					      (region-region el))))))))))))))
      (do-reduce tree))))

(defun dll-only-strings (dll)
  (dll-do (node dll t)
    (let ((el (dll-datum node)))
      (etypecase el
	(string nil)
	(region (return-from dll-only-strings nil))))))

(defun solve ()
  (setf *orig-input* (get-input *standard-input*))
  (handler-case
   (let* ((ifile *orig-input*)
	  (tree (do-parse ifile (length ifile))))
     (setf *best-so-far* tree)
     (setf *orig-input* nil)
     (handler-case (remove-redundancy *best-so-far* (make-props))
       (error (e) (debout "Oops! redund ~A~%" e)))
     (handler-case (reduce-whitespace *best-so-far*)
       (error (e) (debout "Oops! redund ~A~%" e)))
     (load-timer 0)
     (dump tree)
     t)
   (error (e) (load-timer 0) (debout "~A~%" e) (save-day))))

(defun get-number (s)
  (handler-case
   (let* ((*read-eval* nil)
	  (obj (read-from-string s nil nil)))
     (when (numberp obj)
       (floor obj)))
   (error nil nil)))

(defconstant printsecs 3)

(defun main (argv)
  (flet ((usage ()
	   (format *error-output* "Usage: runme <TIME LIMIT>~%")))
    (if (> (length argv) 0)
	(let ((arg (get-number (car argv))))
	  (if (and arg (> arg 0))
	      (progn
		(load-timer (max 1 (- arg printsecs)))
		(solve))
	    (usage)))
      (usage))))

(defun ih-sigalrm (sig code scp)
  (declare (ignore sig code scp))
  (system:with-interrupts
   (debout "Hahaha TIMER!~%")
   (save-day)))

(defun myinit ()
  (system:with-enabled-interrupts ((unix:SIGALRM #'ih-sigalrm))
    (unix:unix-exit (if (main (cddddr ext:*command-line-strings*))
			0
		      1))))
